#
# Created on Tue Apr 16 2024
# Copyright (c) 2024 Huy Truong
# ------------------------------
# Purpose: Evaluator is for supervised, semi-superved, linear prob, fine-tune evaluations
# ------------------------------
#
from typing import Union
from torch.nn.modules import Module
from gentraframe.trainer import Trainer, SingleTrainer
from gentraframe.gen_random_mask_v8 import generate_batch_mask
from gentraframe.utils import *
from gentraframe.config import TrainConfig
from torch_geometric.loader import DataLoader
import torch
from datetime import datetime
import wandb
import time
import numpy as np


class DefaultEvaluator(Trainer):
    def run(self, input_dict: dict[str,Any],  output_keys: list[str], config: TrainConfig) -> Union[dict[str,Any], tuple, Any]:
        task = input_dict['task']
        models = input_dict['models']
        datasets = input_dict['datasets']
        model_path = input_dict['model_paths'][0]
        """perform training or testing"""
        train_args = config

        do_load = config.do_load

        #check conditions
        assert models is not None and len(models) > 0
        assert datasets is not None and len(datasets) >= 1

        #setup some configs
        edge_attrs = train_args.use_data_edge_attrs.split(',') if train_args.use_data_edge_attrs is not None else None
        
        test_loader = DataLoader(datasets[-1], batch_size=train_args.batch_size, shuffle=False) #type:ignore

        device = train_args.device if train_args.device == 'cuda' and torch.cuda.is_available() else 'cpu'
        model_names = [ model.name if hasattr(model,'name') else type(model).__name__ for model in models] #type:ignore
        setattr(train_args, 'model_name', '+'.join(model_names) )
        run_name, postfix = self.start_profiler(task, train_args)
        models = self.load(models, model_path= model_path, device=device, do_load= do_load)
        print('#'*80)
        print(self.as_dict(train_args))
        print('#'*80)
        if task == 'semi':
            print(f'gen-mask version: {generate_batch_mask.__module__}')

        # get loss fn
        criterion = get_criterion(train_args.criterion, device= device)
        
        # gather metrics
        test_metric_fn_dict=  get_metric_fn_collection(prefix='test', task = task) 
       
        #intial records
        best_loss = np.inf
        best_error = np.inf
        best_acc = 0
        best_epoch = 0
        
        start_time = time.time()
        dt1 = datetime.fromtimestamp(start_time)
        print('Start time:', dt1)
        print("*" * 80)

        #############
        #TODO: norm configs should be defined here
        ############
        
        test_loss, test_metric_dict = self.test_one_epoch(models=models,
                                                    loader= test_loader,
                                                    criterion= criterion, 
                                                    metric_fn_dict = test_metric_fn_dict,
                                                    config=train_args,
                                                    task=task,
                                                    )
        
        print_single_metrics(epoch= 0,
                             test_loss=test_loss,
                             test_metric_dict= test_metric_dict)

        if train_args.log_method == 'wandb':
            log_metrics_on_wandb(epoch=0,
                            commit=True,
                            test_loss=test_loss,
                            best_loss=best_loss,
                            best_epoch=best_epoch,
                            test_metric_dict=test_metric_dict,
                            )
                
        end_time = time.time()
        dt2 = datetime.fromtimestamp(end_time) 
        print("*" * 80)
        print('End time:', dt2)
        print('Training time:', dt2 - dt1)

        if train_args.log_method == 'wandb':
            wandb.finish()
        
        return []
    
class SingleEvaluator(DefaultEvaluator, SingleTrainer):
    
    def test_one_epoch(self, models: list[torch.nn.Module],
                        loader: DataLoader,
                        criterion : Callable, 
                        metric_fn_dict: dict[str, Callable],
                        config: TrainConfig,
                        task: str,
                        **kwargs) -> tuple[float, dict]:  
        return DefaultEvaluator.test_one_epoch(self, models, loader, criterion, metric_fn_dict, config, task,  **kwargs)

    def forward_fn(self, models: list[Module], data: Data, batch_mask: Optional[torch.Tensor] = None) -> Any:
        return SingleTrainer.forward_fn(self, models, data, batch_mask)

# class LinearProbingEvaluator(Trainer):

#     @abstractmethod
#     def forward_fn(self, models: list[torch.nn.Module], data: Data, batch_mask: Optional[torch.Tensor] = None) -> Any:
#         pass

#     def get_linear_model(self, in_dim: int, out_dim:int) -> torch.nn.Module:
#         pl: IterPipeline = getattr(self,'pl')
#         pl.pl_config.in_dim = in_dim
#         pl.pl_config.out_dim = out_dim

#         new_model_config = ModelConfig().parse_args()
#         new_model_config.name = 'linear'
#         new_model_config.nc = 1
#         new_model_config.act = None #type:ignore

#         linear_model = pl.model_builder.get_single_model(new_model_config)
#         return linear_model

#     def fit(self, models: list[torch.nn.Module], datasets: list[Dataset], prev_paths: list[str] = []) -> list[str]: 
#         """perform training or testing"""
#         pl_args = self.pl_config
#         args = self.args
#         encoders = models

#         model_path = prev_paths[0]  if len(prev_paths) > 0 else pl_args.model_path 
        
#         task = pl_args.task
#         do_load = True #pl_args.do_load
#         #check conditions
#         assert encoders is not None and len(encoders) > 0
#         assert datasets is not None and (len(datasets) == 2 or len(datasets) ==3)

#         #setup some configs
#         #edge_attrs = args.use_data_edge_attrs.split(',') if args.use_data_edge_attrs is not None else None
#         #use_data_batch = args.use_data_batch
#         device = args.device if args.device == 'cuda' and torch.cuda.is_available() else 'cpu'
#         encoders = self.load(encoders, model_path= model_path, device=device, do_load= do_load)
        
#         [m.eval() for m in encoders]
#         #create embeddings
#         data_dict = {'train': [], 'val': []} if len(datasets) == 2 else  {'train': [], 'val': [], 'test':[]}
#         with torch.no_grad():
#             for key, dataset in zip(list(data_dict.keys()), datasets) :
#                 loader = DataLoader(dataset, batch_size=args.batch_size, shuffle= (key == 'train') ) 
#                 for data in loader:
#                     data.x = data.x.to(device)
#                     data.y = data.y.to(device)
#                     data.edge_index = data.edge_index.to(device)
#                     emb = self.forward_fn(models=encoders, data=data, batch_mask= None)
#                     data.x = emb
#                     if task == 'semi':
#                         data.y = deepcopy(data.x) #data.x.clone()

#                     data_dict[key].append(data)
        
#         #create linear probing loaders
#         in_dim =  data_dict['train'][0].x.shape[-1]
#         out_dim = data_dict['train'][0].y.shape[-1]

#         model = self.get_linear_model(in_dim, out_dim)
#         model.to(device)
        
#         lp_datasets = [data_dict['train'], data_dict['val']] if 'test' not in data_dict else [data_dict['train'], data_dict['val'], data_dict['test']] 

#         self.pl_config.do_load = False
#         return super().fit(models=[model], datasets=lp_datasets) #type:ignore


# class LinearProbingSingleEvaluator(LinearProbingEvaluator, SingleTrainer):
    
#     def forward_fn(self, models: list[Module], data: Data, batch_mask: Optional[torch.Tensor] = None) -> Any:
#         return SingleTrainer.forward_fn(self, models, data, batch_mask)


#
# Created on Wed Jan 23 2024
# Copyright (c) 2024 Huy Truong
# ------------------------------
# Purpose: Store mask generation algorithm
# Version: v8 - random
# Note: It is similar to v7- random, but we add a boolean flag whether we zero required masks from context (i.e., target mask turns off (0) all positions of context that was masked (1))
# ------------------------------
#

import torch
import math
import timeit
from functools import wraps

    
def measure(func, repeat=3, number=10000):
    """Include debug statement and timeit setup"""
    @wraps(func)
    def wrapper(*args,**kwargs):
        res = func(*args,**kwargs)
        times = timeit.repeat(
            stmt=lambda: func(*args,**kwargs),
            repeat=repeat, number=number)
        print(f'The function was run in {repeat} times. The mean execution time is {sum(times)/repeat} seconds.')
        return res
    return wrapper
#@measure
def generate_batch_mask( num_nodes:list[int], mask_rate:float, required_mask: torch.Tensor, do_mask_required_mask=True, **kwargs) -> torch.Tensor:
    assert mask_rate > 0.
    N = sum(num_nodes) 
    #tensor_num_nodes has shape [batch_size]
    cumsum_num_nodes = torch.cumsum(num_nodes,dim=0) #type:ignore
    lbs = torch.nn.functional.pad(cumsum_num_nodes[:-1],[1,0]) #torch.cat([torch.zeros(size=[1]),cumsum_num_nodes[:-1]]) 
    ubs = cumsum_num_nodes
    remaining_nodes = torch.arange(N)

    if required_mask is not None:
        nonzero_inverted_required_mask = (~required_mask).nonzero()
        remaining_nodes = remaining_nodes[nonzero_inverted_required_mask]
       
    def extract_graphs(lb,ub):
        filter_mask = torch.logical_and(remaining_nodes >= lb, remaining_nodes < ub)  
        segment = remaining_nodes[filter_mask]
        gap = segment.shape[0]
        mask_num = max(math.floor(gap * mask_rate),1)
        keep_num = gap - mask_num
        permuted_idx = torch.randperm(segment.nelement())[:keep_num]
        permuted_node = segment[permuted_idx]

        return permuted_node
    
    
    keep_nodes = torch.hstack(tuple(map(extract_graphs, lbs, ubs)))

    mask = torch.ones(N, dtype= torch.bool)
    mask[keep_nodes] = 0
    
    if not do_mask_required_mask:
        mask[required_mask.nonzero()] = 0
    return mask

    

    
    




#
# Created on Sun Mar 31 2024
# Copyright (c) 2024 Huy Truong
# ------------------------------
# Purpose: Simple Trainer
# ------------------------------
#
from typing import Callable, Protocol, Union
from torch.nn.modules import Module
from gentraframe.utils import *
from gentraframe.config import TrainConfig 
from gentraframe.gradient_clipping import GradientClipping
from gentraframe.gen_random_mask_v8 import generate_batch_mask
from torch_geometric.loader import DataLoader
import torch
import torch.nn.functional as F
from datetime import datetime
import wandb
import os
import time
import numpy as np
from torch_geometric.data import Dataset 
from abc import abstractmethod

class Trainer:
    @abstractmethod
    def forward_fn(self, models: list[torch.nn.Module], data: Data, batch_mask: Optional[torch.Tensor] = None) -> Any:
        pass

    def train_one_epoch(self, models: list[torch.nn.Module],
                        optimizer: torch.optim.Optimizer, 
                        loader: DataLoader,
                        criterion : Callable, 
                        metric_fn_dict: dict[str, Callable],
                        config : TrainConfig, 
                        task: str, **kwargs ) -> tuple[float, dict, Any]:  
        edge_attrs = config.use_data_edge_attrs.split(',') if config.use_data_edge_attrs is not None else None
        device = config.device if config.device == 'cuda' and torch.cuda.is_available() else 'cpu'
        mask_rate = config.mask_rate
        use_data_batch = config.use_data_batch
        for pt in models:
            pt.train()
        
        len_loader_dataset = len(loader.dataset) # type:ignore
        total_loss = 0
        total_metric_dict = {k: 0 for k in metric_fn_dict.keys()}

        for data in loader:  # Iterate in batches over the training dataset.
            optimizer.zero_grad()  # Clear gradients.
            
            if task == 'semi':
                num_nodes = torch.unique(data.batch, return_counts=True)[1]
                batch_mask = generate_batch_mask(num_nodes=num_nodes,
                                                edge_index=data.edge_index,
                                                mask_rate=mask_rate,
                                                required_mask=None) #type:ignore
            else:
                batch_mask = data.mask  

            data.x = data.x.to(device)
            data.y = data.y.to(device)
            
            data.edge_attr = data.edge_attr.to(device) if edge_attrs else None
            data.batch = data.batch.to(device) if use_data_batch else None
            data.edge_index = data.edge_index.to(device)

            if task == 'semi':
                data.x[batch_mask] =  0
                out = self.forward_fn(models=models, data=data, batch_mask=batch_mask)
                y_pred = apply_masks(out, [batch_mask]) #out[batch_mask] #type:ignore
                y_true = data.y[batch_mask]
            else:
                data.x = data.x
                out = self.forward_fn(models=models, data=data, batch_mask=batch_mask)
                y_pred = out[batch_mask] if batch_mask is not None else out
                y_true = data.y 


            tr_loss = criterion(y_pred, y_true)
            tr_loss.backward()  # Derive gradients.
            optimizer.step()  # Update parameters based on gradients.

            #forward target
            with torch.no_grad():
                total_loss += float(tr_loss) * data.num_graphs
                y_pred_rescaled = y_pred
                y_true_rescaled = y_true
                
                for k, fn in metric_fn_dict.items():
                    computed_metric = fn(y_pred_rescaled, y_true_rescaled)
                    total_metric_dict[k] += computed_metric * data.num_graphs

        
        with torch.no_grad():
            metric_dict = {k : total_metric_dict[k] / len_loader_dataset for k in total_metric_dict.keys() }
            return total_loss /len_loader_dataset, metric_dict, out
    
    def test_one_epoch(self, models: list[torch.nn.Module],
                        loader: DataLoader,
                        criterion : Callable, 
                        metric_fn_dict: dict[str, Callable],
                        config: TrainConfig,
                        task:str,**kwargs) -> tuple[float, dict]:  
        device = config.device if config.device == 'cuda' and torch.cuda.is_available() else 'cpu'
        edge_attrs = config.use_data_edge_attrs.split(',') if config.use_data_edge_attrs is not None else None
        
        mask_rate = config.mask_rate
        use_data_batch = config.use_data_batch

        for pt in models:
            pt.eval()
        
        with torch.no_grad():
            total_loss = 0
            total_metric_dict = {k: 0 for k in metric_fn_dict.keys()}
            len_loader_dataset = len(loader.dataset) # type:ignore
            for data in loader:
                #assert data.edge_index.max() < data.num_nodes

                if task == 'semi':
                    batch_mask = generate_batch_mask(num_nodes=torch.unique(data.batch, return_counts=True)[1],
                                                    edge_index=data.edge_index,
                                                    mask_rate=mask_rate,
                                                    required_mask=None) #type:ignore
                else:
                    batch_mask = data.mask 

                data.x = data.x.to(device)
                data.y = data.y.to(device)

                data.edge_attr = data.edge_attr.to(device) if edge_attrs else None
                data.batch = data.batch.to(device) if use_data_batch else None
                data.edge_index = data.edge_index.to(device)

                if task == 'semi':
                    data.x[batch_mask] = 0
                    out = self.forward_fn(models=models, data=data, batch_mask=batch_mask)
                    y_pred = apply_masks(out, [batch_mask]) #type:ignore
                    y_true = data.y[batch_mask]
                else:
                    out = self.forward_fn(models=models, data=data, batch_mask=batch_mask)
                    y_pred = out[batch_mask] if batch_mask is not None else out
                    y_true = data.y 

                val_loss = criterion(y_pred, y_true)
                #update metrics
                y_pred_rescaled = y_pred
                y_true_rescaled = y_true
                total_loss += float(val_loss) * data.num_graphs
                for k, fn in metric_fn_dict.items():
                    computed_metric = fn(y_pred_rescaled, y_true_rescaled)
                    total_metric_dict[k] += computed_metric * data.num_graphs

            
            metric_dict = {k : total_metric_dict[k] / len_loader_dataset for k in total_metric_dict.keys() }
            return total_loss / len_loader_dataset, metric_dict

    def as_dict(self, args: TrainConfig) -> dict:

        merged_config = args.as_dict()
        #if self.pl_config is not None:
        #    merged_config.update( self.pl_config.as_dict())
        
        return merged_config
    
    def load(self, models: list[torch.nn.Module], model_path: str, device:str, do_load: bool) -> list[torch.nn.Module] :
        #TODO: model loading should belong to ModelBuilder
        model_names = [ model.name if hasattr(model,'name') else type(model).__name__ for model in models]  #type:ignore
        
        ret_models = []
        for i, m in enumerate(models):
            model : torch.nn.Module = m  #type:ignore
            model_name= model_names[i]
            print(f'#######{model_name}#########')
            print(model)
            print("Model parameters: ", sum(p.numel() for p in model.parameters()))
            
            if do_load:
                if model_path is None :
                    print(f'model_path keys are not found! Please distinguish model path from save_path folder! We use a new model!')
                elif not os.path.exists(model_path):
                    raise FileNotFoundError(f'{model_path} file is not found')
                else:
                    model, cp_dict = load_checkpoint(path=model_path,
                                            model=model,
                                            load_key=getattr(model, 'name'),)
                    
            model = model.to(device, non_blocking=True)
            ret_models.append(model)
        return ret_models

    def start_profiler(self,task: str, args: TrainConfig )-> tuple[str,str]:
        """start wandb
        Returns:
            str: run name
        """
       
        run_prefix= ''

        postfix = datetime.today().strftime('%Y%m%d_%H%M')
        run_name = f'{run_prefix}+{task}+{getattr(args,"model_name")}+{postfix}'
        
        # start a new wandb run to track this script
        if args.log_method == 'wandb':
            wandb.init(
                # set the wandb project where this run will be logged
                project=args.project_name, 
                name = run_name,
                # track hyperparameters and run metadata
                config= self.as_dict(args)
            )

        os.makedirs(args.save_path,exist_ok=True)

        return run_name, postfix
    
    def run(self, input_dict: dict[str,Any],  output_keys: list[str], config: TrainConfig) -> Union[dict[str,Any], tuple, Any]:
        models : list[torch.nn.Module ]= input_dict['models']
        datasets : list[Dataset] = input_dict['datasets']
        do_load = config.do_load
        task= input_dict['task']

        train_args = config

        model_path = config.model_path

        #check conditions
        assert models is not None and len(models) > 0
        assert datasets is not None and len(datasets) == 2

        #setup some configs
        edge_attrs = train_args.use_data_edge_attrs.split(',') if train_args.use_data_edge_attrs is not None else None
        
        train_loader = DataLoader(datasets[0], batch_size=train_args.batch_size, shuffle=True) #type:ignore
        valid_loader = DataLoader(datasets[1], batch_size=train_args.batch_size, shuffle=False) #type:ignore

        device = train_args.device if train_args.device == 'cuda' and torch.cuda.is_available() else 'cpu'
        model_names = [ model.name if hasattr(model,'name') else type(model).__name__ for model in models] #type:ignore
        setattr(train_args, 'model_name', '+'.join(model_names) )
        run_name, postfix = self.start_profiler(task=task, args=train_args)
        models = self.load(models, model_path= model_path, device=device, do_load= do_load)
        print('#'*80)
        print(self.as_dict(train_args))
        print('#'*80)
        if task == 'semi':
            print(f'gen-mask version: {generate_batch_mask.__module__}')

        # get loss fn
        criterion = get_criterion(train_args.criterion, device= device)
        
        # feed models into optimizer
        if len(models) == 1:
            optimizer = torch.optim.Adam(models[0].parameters(), lr=train_args.lr, weight_decay=train_args.weight_decay)
        else:
            param_groups = [{'params': (p for n, p in model.named_parameters())} for model in models]
            optimizer = torch.optim.Adam(param_groups, lr=train_args.lr, weight_decay=train_args.weight_decay) #TODO: add eps
        
        # gather metrics
        train_metric_fn_dict=  get_metric_fn_collection(prefix='train', task = task) 
        val_metric_fn_dict=  get_linear_probing_metric_fn_collection(prefix='val', task = task)
       
        # define scheduler and gradient clipping
        scheduler = None
        grad_clipper = GradientClipping(percentile=10) if train_args.use_gradient_clipping else None
    

        #intial records
        best_loss = np.inf
        best_error = np.inf
        best_acc = 0
        best_epoch = 0
        
        start_time = time.time()
        dt1 = datetime.fromtimestamp(start_time)
        print('Start time:', dt1)
        print("*" * 80)

        #############
        #TODO: norm configs should be defined here
        ############
        best_model_path= ''
        last_model_path= ''
        for epoch in range(1, train_args.epochs + 1):
            tr_loss, tr_metric_dict, out = self.train_one_epoch(models=models,
                                                        optimizer=optimizer , 
                                                        loader=train_loader,
                                                        criterion= criterion, 
                                                        metric_fn_dict = train_metric_fn_dict,
                                                        config=config,
                                                        task =task,
                                                        )
            
            val_loss, val_metric_dict = self.test_one_epoch(models=models,
                                                        loader= valid_loader,
                                                        criterion= criterion, 
                                                        metric_fn_dict = val_metric_fn_dict,
                                                        config = config,
                                                        task =task,
                                                        )
            
            
            if val_loss < best_loss:
                best_loss = val_loss
                best_metric_dict = val_metric_dict
                best_epoch = epoch
                # save training_checkpoint
                best_model_path = os.path.join(config.model_path,f"best_{getattr(train_args,'model_name')}_{postfix}.pth")
                save_kwargs= dict(
                    path= best_model_path,
                    optimizer_state_dict=optimizer.state_dict() if optimizer else None, 
                    epoch=best_epoch, 
                    loss=best_loss, 
                    val_metric_dict = best_metric_dict, 
                    edge_attrs=edge_attrs,
                    norm_type=train_args.norm_type, 
                )

                for model in  models:
                    save_kwargs.update({ getattr(model,'name'):  model.state_dict()})
                save_checkpoint(**save_kwargs) #type:ignore

            if (epoch == 1 or (epoch % train_args.log_per_epoch) == 0 ) and not math.isnan(tr_loss):
                print_metrics(epoch=epoch,
                        tr_loss=tr_loss, 
                        val_loss=val_loss,
                        tr_metric_dict=tr_metric_dict, 
                        val_metric_dict=val_metric_dict,
                        )
                last_model_path=os.path.join(config.model_path,f"last_{getattr(train_args,'model_name')}_{postfix}.pth")
                save_kwargs= dict(
                    path= last_model_path,
                    optimizer_state_dict= optimizer.state_dict() ,   #type:ignore
                    epoch=best_epoch, #type:ignore
                    loss=best_loss,   #type:ignore
                    val_metric_dict = best_metric_dict, 
                    edge_attrs=edge_attrs,
                    norm_type=train_args.norm_type, 
                )
                for model in  models:
                    save_kwargs.update({ getattr(model,'name'):  model.state_dict()})
                save_checkpoint(**save_kwargs) #type:ignore

            if train_args.log_method == 'wandb':
                log_metrics_on_wandb(epoch=epoch,
                                commit=True,
                                train_loss=tr_loss,
                                val_loss=val_loss,
                                best_loss=best_loss,
                                best_epoch=best_epoch,
                                tr_metric_dict=tr_metric_dict,
                                val_metric_dict=val_metric_dict,
                                )
                
            
            if scheduler is not None:
                scheduler.step(val_loss)  
        
        end_time = time.time()
        dt2 = datetime.fromtimestamp(end_time) 
        print("*" * 80)
        print('End time:', dt2)
        print('Executation time: ', dt2 - dt1)

        if train_args.log_method == 'wandb':
            wandb.finish()
        
        if best_model_path != '':
            model_paths  = [best_model_path]
        else:
            model_paths = [last_model_path]

        #input_dict['model_paths'] = model_paths
        return model_paths

class SingleTrainer(Trainer):
    def forward_fn(self, models: list[Module], data: Data, batch_mask: Optional[torch.Tensor] = None) -> Any:
        assert len(models) == 1
        out =  models[0](x=data.x, edge_index=data.edge_index, batch=data.batch, edge_attr=data.edge_attr)
        return out
    



#
# Created on Fri May 31 2024
# Copyright (c) 2024 Huy Truong
# ------------------------------
# Purpose: A generic pypeline
# ------------------------------
#

from typing import Protocol, Any, Optional
import networkx as nx
import matplotlib.pyplot as plt
from gentraframe.config import GTConfig 
from abc import abstractmethod
import inspect
from gentraframe.utils import RunFn


class ModuleV2(Protocol):
    @abstractmethod
    def get_inpout_config(self)-> Optional[GTConfig]: 
        """return the input output config"""
        pass
       
    
class ComponentV2():
    def __init__(self, name: str, inputs: list[str], outputs: list[str], module: ModuleV2) -> None:
        self.name = name
        self.inputs = inputs
        self.outputs = outputs
        self.module = module
        

    def attach(self, parent: Any) -> tuple[str, list[str], list[str]]:
        self.parent = parent
        return self.name, self.inputs, self.outputs
    
    def run(self, input_dict: dict[str,Any], **kwargs) -> dict[str,Any]: #type:ignore
        """ execute the module
        """
        out_dict = {}
        #get current inpout config
        inpout_config = self.module.get_inpout_config()
        
        #perform module.run        
        #run_fn = self.module.get_run()
        
        assert hasattr(self.module, 'run')
        run_fn: RunFn = getattr(self.module,'run')

        parameters  =  inspect.signature(run_fn).parameters
        assert 'inpout' in parameters, f'inpout is not in the parameters of function {run_fn}'
        inpout_param : inspect.Parameter = parameters['inpout']
        annotation = inpout_param.annotation
        if annotation == inspect.Parameter.empty:
            print('WARNING! param: inpout has to be type defined! Type is unsafe!')
            casted_inpout_config = inpout_config
            out = run_fn(casted_inpout_config, self.outputs, **kwargs)  #type:ignore
        else:
            casted_inpout_config = None
            if inpout_config is not None:
                print(f'annotation = {annotation}')
                print(f'type(annotation) = {type(annotation)}')
                casted_inpout_config = annotation.__origin__(inpout_config)()
                #add values into inpout
                for k in self.inputs:
                    setattr(casted_inpout_config, k, input_dict[k])
            out = run_fn(casted_inpout_config, self.outputs, **kwargs)  
        

        #check attributes and assing values into output dict
        if isinstance(out, dict):
            assert set(self.outputs).intersection(out)
            out_dict = out
        elif isinstance(out, tuple):
            assert len(self.outputs) == len(out)
            out_dict = {o_key: out[i]  for i,o_key in enumerate(self.outputs)}
        else:
            if len(self.outputs) > 0:
                assert len(self.outputs) == 1
                out_dict = {self.outputs[0]: out}
            else:
                out_dict = {}
        #merge the existing values from input_dict
        out_dict.update(input_dict)
        return out_dict


class Pipeline(ComponentV2):
    def __init__(self, name: str, inputs: list[str]=[], outputs: list[str]=[], com_list : list[ComponentV2] = [], verbose: bool=True) -> None:
        self.name = name
        self.inputs = inputs
        self.outputs = outputs
        ##
        self.verbose = verbose
        self.com_dict : dict[str, ComponentV2] = {}
        self.in_dict  = {}
        self.out_dict = {}
        self.name_roots : list[str] = []
        self.g = nx.DiGraph()
        for com in com_list:
            name, inputs, outputs = com.attach(self)
            assert name not in self.com_dict, f'error: name {name} is duplicated! In context of pipeline {self.name}, all names have to be unique!'
            self.com_dict[name] = com 
            self.in_dict[name] = inputs
            self.out_dict[name] = outputs
            self.g.add_node(name, visited=False)

            if len(inputs) == 0:
                self.name_roots.append(name)
        
        
        for i in range(len(com_list)-1):
            i_node  = com_list[i].name
            i_in_nodes = self.in_dict[i_node]
            i_out_nodes = self.out_dict[i_node]
            for j in range(i,len(com_list)):
                j_node  = com_list[j].name
                j_in_nodes = self.in_dict[j_node]
                j_out_nodes = self.out_dict[j_node]

                edge_label = set(i_out_nodes).intersection(j_in_nodes)
                if edge_label:
                    self.g.add_edge(i_node, j_node, edge_label=edge_label)
                
                
                edge_label =set(j_out_nodes).intersection(i_in_nodes)
                if edge_label:
                    self.g.add_edge(j_node, i_node, edge_label=edge_label)
        
        if len(self.name_roots) == 0:
            self.name_roots = [node_name for node_name in self.g.nodes if len(self.g.in_edges(node_name)) == 0]

        if self.verbose:
            nx.draw(self.g, with_labels=True)
            pos = nx.spring_layout(self.g)
            nx.draw_networkx_edge_labels(
                self.g, pos,
                font_color='red',
            )
            plt.show(block=True)
        

    def reset(self):
        for _, node_data in self.g.nodes(data=True): 
            node_data['visited'] = False

    def process_node(self, name: str, next_dict: dict, **kwargs) -> dict:
        if not self.g.nodes[name]['visited']:
            assert set(next_dict.keys()).issuperset(self.in_dict[name])
            next_dict = self.com_dict[name].run(next_dict,   **kwargs)
            assert isinstance(self, Pipeline) or set(next_dict.keys()).intersection(self.out_dict[name]), f'Expected output keys: {self.out_dict[name]}, but get output keys= {next_dict.keys()}'
            self.g.nodes[name]['visited'] = True
        return next_dict

    def run(self, input_dict: dict[str,Any], **kwargs) -> dict[str,Any]: #type:ignore
        next_dict = input_dict
        for root in self.name_roots:
            next_dict = self.process_node(root, next_dict,  **kwargs)
        
        for root in self.name_roots:
            nodes = nx.traversal.bfs_tree(self.g, source = root)
            for node in nodes:
                next_dict = self.process_node(node, next_dict, **kwargs)

        return next_dict
    


Component = ComponentV2
Module = ModuleV2


#
# Created on Fri May 31 2024
# Copyright (c) 2024 Huy Truong
# ------------------------------
# Purpose: 
# ------------------------------
#
from typing import Literal, Any
from tap import Tap
import yaml
import os

class GTConfig(Tap):
    
    """An abstract config for inheritance
    """
    def _from_yaml(self, yaml_path, skip_unsettable=False)->None:
        assert yaml_path[-4:] == 'yaml'
        assert os.path.isfile(yaml_path)

        with open(yaml_path, 'r') as yaml_in:
            yaml_object = yaml.unsafe_load(yaml_in) #yaml.safe_load(yaml_in) # yaml_object will be a list or a dict
            
            self.from_dict(yaml_object,skip_unsettable=skip_unsettable)

    def _to_yaml(self, yaml_path)->dict:
        from collections import OrderedDict
        assert yaml_path[-4:] == 'yaml' 

        kv_dict = self.as_dict()
        yaml_object = OrderedDict()
        for cv in self.class_variables:
            assert cv in kv_dict, f'class_variables has a key {cv} is not found in kv_dict'
            yaml_object[cv] = kv_dict[cv]

        #Order reservation in YAML file# ref: https://stackoverflow.com/questions/45253643/order-preservation-in-yaml-via-python-script    
        def ordered_dict_representer(self, value): 
            return self.represent_mapping('tag:yaml.org,2002:map', value.items())
        
        yaml.add_representer(OrderedDict, ordered_dict_representer)
        with open(yaml_path,'w') as yaml_out:
            yaml.dump(yaml_object, yaml_out, indent = 4, sort_keys=False)

        return yaml_object
    

# class PipelineConfig(GTConfig):
#     task: Literal['semi','supervised'] # task we work with
#     node_in_dim: int = -1 # node in dim
#     edge_in_dim: int = -1 # edge in dim
#     out_dim: int = -1 # label out dim

class ModelConfig(GTConfig):
    name : str = 'gcn' #name of the model
    num_layers: int = 2 #number of layers
    nc: int = 16 #number of hidden nc
    act: Literal['relu','gelu'] = 'relu' #activation
    has_final_linear: bool = False # if true, add a linear following the gnn layers
    weight_path: str # path storing the model weights
    do_load: bool # load weights of the model 

class TrainConfig(GTConfig):
    model_path: str = '' #temp
    do_load: bool = False # flag represents temporarily global setting for all mdoels
    model_configs: list[ModelConfig] = [] # involving DL models in this train
    lr: float= 0.01 #0.01 #Learning rate. Default is 0.0005
    weight_decay: float= 5e-4 #weight decay. Default is 0.000006
    epochs: int= 100 #number of epochs to train the model
    mask_rate: float=0.95 #masking ratio. Default is 0.95
    criterion: Literal["mse", "mae", "sce","ce"] ='ce' # criterion loss. Support mse|sce|mae
    batch_size: int= 64 #batch size
    use_data_batch: bool=False #pass pyg data batch as parameter into model. Set False to fasten training. Default is False
    use_data_edge_attrs: str= None # pass pyg data edge attributes. Eg: [diameter,length] for multi-use. Support: diameter| length| None #type:ignore
    device:str='cuda' #Training device. If gpu is unavailable, device is set to cpu. 
    norm_type: Literal["znorm", "minmax", "unused"]='unused' #normalization type. Support znorm| minmax|unused"
    use_gradient_clipping=False # flag indicating whether using gradient clipping.
    """###########################TRACKING EXPERIMENTS SETTINGS################################"""
    log_method: str = '' #log method! Support wandb or None
    #log_gradient: bool = False #flag indicates keeping track of gradient flow
    project_name: str ='dev-pretext-train' #name of tracking project"
    save_path: str = 'experiments_logs/test_args/fun_test' #Path to store model weights
    log_per_epoch: int = 1# log every log_per_epoch 
    run_prefix : str = ''# it helps naming the run on WANDB
    
    """#########################################################################################"""
    
    def configure(self) -> None:
        def str_or_none(string: str) -> str:
            return string if string is not None else None # type:ignore
        self.add_argument('--use_data_edge_attrs',type=str_or_none)
    




import numpy as np
import torch
from typing import Union
def get_grad_norm(model):
    """
    Adapted from AutoClip: Adaptive Gradient Clipping
    Prem Seetharaman, Gordon Wichern, Bryan Pardo, Jonathan Le Roux. "AutoClip: Adaptive Gradient
    Clipping for Source Separation Networks." 2020 IEEE 30th International Workshop on Machine Learning
    for Signal Processing (MLSP). IEEE, 2020.

    https://github.com/pseeth/autoclip
    """
    total_norm = 0
    for p in model.parameters():
        if p.grad is not None:
            param_norm = p.grad.data.norm(2)
            total_norm += param_norm.item() ** 2
    total_norm = total_norm ** (1. / 2)
    return total_norm

def get_grad_norm_for_models(models: list):
    """
    Adapted from AutoClip: Adaptive Gradient Clipping
    Prem Seetharaman, Gordon Wichern, Bryan Pardo, Jonathan Le Roux. "AutoClip: Adaptive Gradient
    Clipping for Source Separation Networks." 2020 IEEE 30th International Workshop on Machine Learning
    for Signal Processing (MLSP). IEEE, 2020.

    https://github.com/pseeth/autoclip
    """
    total_norm = 0
    for model in models:
        for p in model.parameters():
            if p.grad is not None:
                param_norm = p.grad.data.norm(2)
                total_norm += param_norm.item() ** 2
    total_norm = total_norm ** (1. / 2)
    return total_norm

class GradientClipping:
    def __init__(self, percentile: float =10) -> None:
        self.grad_history = []
        self.percentile = percentile

    def reset(self):
        self.grad_history.clear()

    def cache_gradient_and_compute_norm(self, model: Union[torch.nn.Module,list]) -> float:
        if isinstance(model, list):
            self.grad_history.append(get_grad_norm_for_models(model))
        else:
            self.grad_history.append(get_grad_norm(model))
        clip_value = np.percentile(self.grad_history, self.percentile)
        return clip_value   
    
    def cache_gradient_norm(self, norm) -> float:
        self.grad_history.append(norm)
        clip_value = np.percentile(self.grad_history, self.percentile)
        return clip_value


#
# Created on Fri May 31 2024
# Copyright (c) 2024 Huy Truong
# ------------------------------
# Purpose: Store problems
# ------------------------------
#
from abc import ABC, abstractmethod
from typing import Literal, Optional, Any

class Experiment(ABC):
    task : Literal['supervised','semi']

    @abstractmethod
    def get_dimensions(self, axis: int = -1) -> tuple[int, Optional[int], int]:
        pass

    @abstractmethod
    def get_data(self, *args,**kwargs)-> tuple[Any, Any, Any]:
        pass

    


#
# Created on Sun Mar 31 2024
# Copyright (c) 2024 Huy Truong
# ------------------------------
# Purpose: Store general functions
# ------------------------------

import numpy as np
from copy import deepcopy
import torch
from functools import partial
import torch.nn.functional as F
from typing import Any, Callable, Literal, Optional, Union, Protocol, Type
import torch_geometric
from torch_geometric.data import Data
import wandb
import math
import networkx as nx
import torch_geometric.utils as pgu
import timeit
from functools import wraps
from gentraframe.config import GTConfig

class RunFn(Protocol):
    def __call__(self, inpout: Any, outputs : list[str], **kwargs) -> Union[dict[str,Any], tuple, Any]:
        pass

def scale(data: Any, norm_type: str="minmax", mean: Any=None, std: Any=None, min: Any=None, max: Any=None, eps: float=1e-8)->Any:
    """scale function supports normalization

    Args:
        data (Any): unnormed data
        norm_type (str, optional): minmax or znorm . Defaults to "minmax".
        mean (Any, optional): mean values. Defaults to None.
        std (Any, optional): std values. Defaults to None.
        min (Any, optional): min values. Defaults to None.
        max (Any, optional): max values. Defaults to None.
        eps (float, optional): a very small number to avoid zero division. Defaults to 1e-8.

    Returns:
        Any: normed data
    """
    assert norm_type in ['minmax', "znorm"]
    if norm_type == "minmax":
        assert min is not None  and max is not None , "min and max values are missing"
        return (data - min) / (max - min)
    elif norm_type == "znorm":
        assert mean is not None and std is not None , "mean and std values are missing"
        return (data - mean) / (std + eps)
    

def descale(scaled_data: Any, norm_type:str="minmax", mean:Any=None, std:Any=None, min:Any=None, max:Any=None)->Any:
    """Descale function supports denormalization

    Args:
        scaled_data (Any): normed data
        norm_type (str, optional): znorm or minmax. Defaults to "minmax".
        mean (Any, optional): mean values. Defaults to None.
        std (Any, optional): std values. Defaults to None.
        min (Any, optional): min values. Defaults to None.
        max (Any, optional): max values. Defaults to None.

    Returns:
        Any: unnormed data
    """
    if norm_type == 'minmax':
        assert min  is not None and max is not None, "min and max values are missing"
        data = (scaled_data * (max-min)) + min
    elif norm_type == 'znorm':
        assert mean is not None and std is not None, "mean and std values are missing"
        data = (scaled_data * std) + mean
    else:
        data = scaled_data
    return data

    
def nx_to_pyg(data: Any, graph: Data)-> Data:
    """ load data import graph Data containing topology
    Args:
        data (Any): 
        graph (Data): Data converted from nx graph containing topology

    Returns:
        torch_geometric.data.Data: pyg data format
    """
    g_data = deepcopy(graph)
    y = data
    g_data.y = torch.Tensor(np.reshape(y, [-1, 1])) #type:ignore
    g_data.x = torch.Tensor(np.reshape(y, [-1, 1])) #type:ignore
    return g_data

def calculate_classified_accuracy(y_pred, y_true):
    y_true = y_true.max(1)[1].long()#.squeeze().long()
    preds = y_pred.max(1)[1].type_as(y_true)
    correct = preds.eq(y_true).double()
    correct = correct.sum().item()
    return correct / len(y_true) 

def calculate_nse(y_pred, y_true, exponent=2):
    raveled_y_pred = torch.ravel(y_pred) 
    raveled_y_true = torch.ravel(y_true) 
    return 1.0 - torch.div(torch.sum(torch.pow(raveled_y_pred - raveled_y_true, exponent)) , torch.sum(torch.pow(raveled_y_true - torch.mean(raveled_y_true), exponent)) + 1e-12) 

def calculate_rmse(y_pred, y_true):
    return torch.sqrt(torch.mean((y_pred-y_true)**2))

def calculate_rel_error(y_pred, y_true, epsilon: float =1e-6):
    err = torch.abs(torch.subtract(y_true, y_pred))
    mask = (torch.abs(y_true) > 0.01)
    rel_err = torch.abs(torch.divide(err[mask], y_true[mask] + epsilon))
    return torch.mean(rel_err)

def calculate_rel_error2(y_pred, y_true, epsilon: float =1e-4):
    flat_pred = y_pred.reshape([-1,1])
    flat_true = y_true.reshape([-1,1])
    err = torch.abs(torch.subtract(flat_true, flat_pred))
    mask = (torch.abs(flat_true) > 0.01)
    rel_err = torch.abs(torch.divide(err[mask], flat_true[mask] + epsilon))
    return torch.mean(rel_err)


def calculate_accuracy(y_pred, y_true, threshold=0.2):
    mae = torch.abs(torch.subtract(y_true, y_pred))
    acc = (mae <= (y_true * threshold)).float()
    return torch.mean(acc)


def calculate_correlation_coefficient(y_pred, y_true):
    vx = y_pred - torch.mean(y_pred)
    vy = y_true - torch.mean(y_true)

    cost = torch.sum(vx * vy) / (torch.sqrt(torch.sum(vx ** 2)) * torch.sqrt(torch.sum(vy ** 2)))
    #cov    = torch.mul(y_pred-y_pred.mean(), y_true-y_true.mean()).mean()
    #std   = torch.sqrt(torch.mul(torch.square(y_pred-y_pred.mean()), torch.square(y_true-y_true.mean()))).mean()
    
    return torch.clamp(cost, -1.0, 1.0)

def calculate_r2(y_pred, y_true):
    r = calculate_correlation_coefficient(y_pred, y_true)
    return r**2

def inverse_mask_tuple(mask_tuple : tuple) -> tuple:
    inverse_tuple = [~x for x in mask_tuple]
    return tuple(inverse_tuple)

def apply_masks(x: torch.Tensor, masks: list[torch.Tensor]):
    # x has shape (num_masks, batch_size * num_graphs, num_features)
    # masks has length = num_masks, masks's element has shape  (batch_size* num_graphs)

    num_masks = len(masks)
    if num_masks == 1:
        if len(x.size()) == 2:
            return x[masks[0]]
        else:
            return x.squeeze(0)[masks[0]]


    xs=[]
    for i in range(num_masks):
        mask = masks[i]
        xs.append(x[i][mask])

    # out has shape = (num_masks, batch_size * num_graphs, num_features)
    return torch.cat(xs,dim=0)


def get_metric_fn_collection(prefix:str, task: Literal['supervised','semi'] = 'semi')->dict:
    """util creating metric funtions

    Args:
        prefix (str): set a prefix name for tracking these experiment

    Returns:
        dict: contains functional name and callable functions
    """
    if task == 'semi':
        metric_fn_dict= {
            f'{prefix}_error': calculate_rel_error2,
            f'{prefix}_0.1': partial(calculate_accuracy,threshold=0.1),
            f'{prefix}_corr': calculate_correlation_coefficient,
            f'{prefix}_r2': calculate_r2,
            f'{prefix}_mae': F.l1_loss,
            f'{prefix}_rmse': calculate_rmse,
            f'{prefix}_mynse': partial(calculate_nse,exponent=2),
        }
    else:
        metric_fn_dict= {
            f'{prefix}_acc': calculate_classified_accuracy,
        }
    return metric_fn_dict

def get_linear_probing_metric_fn_collection(prefix:str, task: Literal['supervised','semi'] = 'semi')->dict:
    """util creating metric funtions

    Args:
        prefix (str): set a prefix name for tracking these experiment

    Returns:
        dict: contains functional name and callable functions
    """
    if task == 'semi':
        metric_fn_dict= {
            f'{prefix}_error': calculate_rel_error2,
            f'{prefix}_mae': F.l1_loss,
            f'{prefix}_rmse': calculate_rmse,
        }
    else:
        metric_fn_dict= {
            f'{prefix}_acc': calculate_classified_accuracy,
        }
    return metric_fn_dict

def load_checkpoint(path: str, model: torch.nn.Module, load_key:str ='model_state_dict') -> tuple[torch.nn.Module, dict]:
    """support load model and relevant data

    Args:
        path (str): checkpoint file
        model (torch.nn.Module): a model architecture to load weights into
        load_key (str): a key indicate the corresponding model weights

    Returns:
        tuple[torch.nn.Module, dict]: tuple of loaded model and relevant data as dict
    """
    models, cp_dict = load_custom_checkpoint(path, models= [model], load_keys=[load_key])
    return models[0], cp_dict

def load_custom_checkpoint(path: str, models: list[torch.nn.Module], load_keys: list[str]) -> tuple[list[torch.nn.Module], dict]:
    """support load multiple models and relevant data

    Args:
        path (str): checkpoint file
        models (list[torch.nn.Module]): model architectures to load weights into
        load_keys (list[str]): key list indicate the corresponding model weights

    Returns:
        tuple[list[torch.nn.Module], dict]: tuple of list of loaded model and relevant data as dict
    """
    assert path[-4:] == '.pth'
    assert models is not None
    assert len(load_keys) == len(models)
    cp_dict = torch.load(path)
    for i in range(len(models)):
        name = models[i].name if hasattr(models[i],'name') else type(models[i]).__name__
        if load_keys[i] in cp_dict:
            try:
                models[i].load_state_dict(cp_dict[load_keys[i]])
                print(f'Model at position {i}...loaded')
            except RuntimeError:
                print(f'Unable to load weights from {load_keys[i]} to model {name}! Use new model instead!')
        else:
            print(f'WARNING! models at position {i} has no state dict! Use new model instead!')
    return models, cp_dict

def save_checkpoint(path:str,**kwargs) ->str:
    """support save checkpoint. User can leverage kwargs to store model and relevant data

    Args:
        path (str): saved path

    Returns:
        str: saved path
    """
    torch.save(kwargs,path)
    return path

def print_single_metrics(epoch:int, **kwargs):
    formatter =f'Epoch: {epoch:03}, ' # f'Epoch: {epoch:0.3d}'
    for k, v in kwargs.items():
        if isinstance(v, dict):
            for sk, sv in v.items():
                formatter +=  f'{sk}: {sv:.4f}, '
        else:
            formatter += f'{k}: {v:.4f}, '
    print(formatter)

def print_metrics(epoch:int,tr_loss:float,val_loss:float, tr_metric_dict:dict, val_metric_dict:dict):
    """support pretty print string format

    Args:
        epoch (int): current epoch/ trial
        tr_loss (float): training loss
        val_loss (float): validation loss
        tr_metric_dict (dict): training metric dict including name, values
        val_metric_dict (dict): validation metric dict including name, values
    """
    metric_log = ''

    for k, v in tr_metric_dict.items():
        metric_log += f'{k}: {v:.4f}, '
    
    if val_metric_dict:
        for k, v in val_metric_dict.items():
            metric_log += f'{k}: {v:.4f}, '
    
    formatter =  f"Epoch: {epoch:03d}, train loss: {tr_loss:.4f},"
    formatter += f"val_loss: {val_loss:.4f}," if val_loss else ''
    formatter += f" {metric_log}"
    print(formatter)

def print_multitest_metrics(trials:int,mean_test_loss:float, std_test_loss:float,mean_test_sensor_loss:float,std_test_sensor_loss:float,out_test_metric_dict:dict, out_test_sensor_metric_dict:dict ):
    """support beautifying string format for multi-trial evaluation

    Args:
        trials (int): number of trials
        mean_test_loss (float): mean of test losses
        std_test_loss (float): std of test losses
        mean_test_sensor_loss (float): mean of test sensor losses
        std_test_sensor_loss (float): std of test sensor losses
        out_test_metric_dict (dict): metric dict contains mean, std and raw measurment values in trials runs
        out_test_sensor_metric_dict (dict): metric dict contains mean, std and raw measurment sensor values in trials runs
    """
    metric_log = ''
    for k, v in out_test_metric_dict.items():
        if 'mean' in k:
                name = k[:-5]
                std = out_test_metric_dict[f'{name}_std']
                metric_log += f'{name}: {v:.4f} +/- {std:.4f}, '

    for k, v in out_test_sensor_metric_dict.items():
        if 'mean' in k:
                name = k[:-5]
                std = out_test_sensor_metric_dict[f'{name}_std']
                metric_log += f'sensor_{name}: {v:.4f} +/- {std:.4f}, '

    print(f'\nThis TEST experiment reports the average result of {trials} runs.') 
    print(f"test_loss: {mean_test_loss:.4f} +/- {std_test_loss:.4f}, sensor_test_loss: {mean_test_sensor_loss:.4f} +/- {std_test_sensor_loss:.4f}, {metric_log}")
    
def get_gradient_norm(model: torch.nn.Module,norm_type = 2)-> tuple[torch.Tensor, list[torch.Tensor], list[str]]:
    """support get gradient norms from a model. Tracking modules whose names has 'block', 'mlp' or 'res' words

    Args:
        model (torch.nn.Module): any torch module 
        norm_type (int, optional): calculated type of norm. Defaults to 2.

    Returns:
        tuple[torch.Tensor, list[torch.Tensor], list[str]]: returned sum norms of all blocks, list of each norms, list of each name
    """
    block_norms = []
    block_names = []
    for name,param in model.named_parameters():
        # if 'block' in name or 'mlp' in name or 'res' in name :
        #     block_norms.append(torch.norm(param.grad.detach() , p=norm_type) ) 
        #     block_names.append(name)

        if (param.grad is not None) and not (name.endswith('.bias') or len(param.shape) == 1):
            block_norms.append(torch.norm(param.grad.data, p=norm_type))
            block_names.append(name)


    total_norm =  torch.norm(torch.stack(block_norms), norm_type)
    return total_norm, block_norms,block_names

def log_metrics_on_wandb(epoch:int,commit:bool=True, is_epoch_a_trial=False, **kwargs):
    """support function allowing to push log to wandb server

    Args:
        epoch (int): deterministic epoch
        commit (bool, optional): if it is one of non-last incremental logs, set it to True. Defaults to True.
    """
    for k,v in kwargs.items():
        if isinstance(v,dict):
            wandb.log(v,commit=False)
        else:
            wandb.log({k:v},commit=False)
    if is_epoch_a_trial:
        wandb.log({'trial': epoch},commit=commit)
    else:
        wandb.log({'epoch': epoch},commit=commit)

def create_random_graph(num_nodes:int, edge_form_prob:float = 0.5, include_pos:bool=True, radius= 10, graph_id = 0) -> Data: 
    G = nx.gnp_random_graph(num_nodes, edge_form_prob, directed=False) 
    test_features = {}
    pos_features = {}
    g_features = {}
    # sign = 1
    # for i, node_name in enumerate(G.nodes()): 
    #     test_features[node_name] = i
    #     pos_features[node_name] = (i, sign * i)
    #     sign = -1 * sign

    r = radius
    horizontal_distance_between_graphs = graph_id * 4*radius
    for i, node_name in enumerate(G.nodes()): 
        test_features[node_name] = i 
        pos_features[node_name] = (r * np.cos(i/num_nodes * 2* np.pi) + horizontal_distance_between_graphs, r * np.sin(i/num_nodes * 2* np.pi)) if i !=0 else (horizontal_distance_between_graphs,0)
        g_features[node_name] = graph_id

    nx.set_node_attributes(G, test_features, name='x')
    nx.set_node_attributes(G, g_features, name='graph_id')
    if include_pos:
        nx.set_node_attributes(G, pos_features, name='pos') 
    return pgu.from_networkx(G)

    
def measure(func, repeat=3, number=10000):
    """Include debug statement and timeit setup"""
    @wraps(func)
    def wrapper(*args,**kwargs):
        res = func(*args,**kwargs)
        times = timeit.repeat(
            stmt=lambda: func(*args,**kwargs),
            repeat=repeat, number=number)
        print(f'The function was run in {repeat} times. The mean execution time is {sum(times)/repeat} seconds.')
        return res

def load_secret(secret_path:str) -> Optional[dict]:
    if secret_path is None:
        return None 
    secret_extension = secret_path[-2:]
    if secret_extension == 'py':
        import importlib.util
        import sys
        spec = importlib.util.spec_from_file_location("mysecret", secret_path)
        module = importlib.util.module_from_spec(spec)#type:ignore
        sys.modules["mysecret"] = module 
        spec.loader.exec_module(module) #type:ignore
        return module.secrets
    else:
        raise NotImplementedError()
    
def get_criterion(loss_name: str, device: str) -> Callable:
    """ get loss by name """
    if loss_name == 'sce':
        def sce_loss(x, y, alpha = 3):
            x = F.normalize(x,p=2,dim=-1)
            y = F.normalize(y,p=2,dim=-1)
            loss = (1.0 - (x*y).sum(dim=-1)).pow_(alpha)
            return loss.mean()

        criterion = sce_loss
    elif loss_name == 'mse':
        criterion = torch.nn.MSELoss(reduction="mean").to(device)
    elif loss_name == 'mae':
        criterion = torch.nn.L1Loss(reduction="mean").to(device)
    elif loss_name == 'ce':
        criterion =torch.nn.CrossEntropyLoss(reduction="mean").to(device)
    else:
        raise KeyError(f'criterion {loss_name} is not supported')

    return criterion


#
# Created on Fri Apr 26 2024
# Copyright (c) 2024 Huy Truong
# ------------------------------
# Purpose: create a module for pipeline
# ------------------------------
from copy import deepcopy
from typing import  Any, Literal, Optional, Union
from torch import nn
from torch_geometric.data import Data, Dataset
from torch_geometric.datasets.planetoid import Planetoid
from torch_geometric.transforms import NormalizeFeatures
import torch_geometric.utils as pgu
import torch.nn.functional as F
import torch
from torch_geometric.nn import GCN, MLP, GCNConv
from gentraframe.config import ModelConfig, TrainConfig




class CustomDataset(Dataset):
    def __init__(self, data_list):
        super(CustomDataset, self).__init__()
        self.data_list = data_list

    def len(self):
        return len(self.data_list)

    def get(self, idx):
        return self.data_list[idx]
    
class MyCora3(Planetoid):

    def __init__(self, 
                 do_scale:bool,
                 root:str, 
                 remove_isolated:bool,
                 ) -> None:
        
        norm_features_tf = NormalizeFeatures() if do_scale else None
        super().__init__(root=root, 
                         name='Cora',
                         #split='public', 
                         transform= norm_features_tf)
        self.remove_isolated = remove_isolated
        
    def get_set(self, from_set:Literal['train','val','test'], task: Literal['semi','super']) ->  Dataset:
        mask_name = f'{from_set}_mask'
        data = self.get(0)
        assert hasattr(data, mask_name)
        mask = getattr(data, mask_name)
        new_data = deepcopy(data)
        if task == 'semi':
            new_data.x = new_data.x[mask]
            new_data.y =  deepcopy(new_data.x)
        else:
            new_data.x = new_data.x 
            new_data.y = F.one_hot(new_data.y[mask], self.num_classes).to(torch.float32) # new_data.y[mask] # new_data.y[mask] #
            new_data.mask = mask
            
        if self.remove_isolated:
            keep_indices = pgu.mask_to_index(mask)
            new_data.edge_index, new_data.edge_attr=  pgu.subgraph(keep_indices,edge_index=new_data.edge_index, edge_attr= new_data.edge_attr, relabel_nodes=True, return_edge_mask=False) #type:ignore
            
        return CustomDataset([new_data])
      
class CoraProvider:
    def run(self, input_dict: dict[str,Any], output_keys: list[str], **kwargs)-> Union[dict[str,Any], tuple, Any]:
        task = input_dict['task']
        from_sets = input_dict['from_sets']
        self.my_cora = MyCora3( do_scale=True, root= 'cora', remove_isolated= (task == 'semi')) 
        #communicate with pipeline config
        sample : Data =   self.my_cora.get_set(from_set=from_sets[0], task= task).get(0) #type:ignore
        in_dim = sample.x.shape[-1]
        out_dim = sample.y.shape[-1]
        print(sample)

        input_dict['in_dim'] = in_dim
        input_dict['out_dim'] = out_dim
        input_dict['datasets'] = [self.my_cora.get_set(from_set=from_set, task=task)  for from_set in from_sets ] #type:ignore

        return input_dict
    
    
class DefaultModelBuilder:
    def _get_single_model(self, config : ModelConfig, in_dim: int, out_dim:int)->nn.Module:
        name = config.name
       
        assert in_dim >0 and out_dim >0
        
        if name == 'gcn':
            return GCN(hidden_channels=config.nc,
                       num_layers=config.num_layers,
                       in_channels=in_dim,
                       out_channels=out_dim,
                       act= config.act,
                       act_first=True,
                       )
        elif name in ['ffn','mlp']:
            return MLP(hidden_channels=config.nc,
                       num_layers=config.num_layers,
                       in_channels=in_dim,
                       out_channels=out_dim,
                       act= config.act,
                       norm = None,
                       )
        elif name =='linear':
            return MLP(hidden_channels=config.nc,
                       num_layers=1,
                       in_channels=in_dim,
                       out_channels=out_dim,
                       act= None,
                       norm = None,
                       )
        else:
            raise NotImplementedError(f'{name} is not supported.')

    def get_model_by_config(self, model_config: ModelConfig, in_dim : int, out_dim: int) -> nn.Module:
        model = self._get_single_model(model_config,in_dim= in_dim, out_dim= out_dim)
        setattr(model, 'name', model_config.name)
        return model

    def run(self, input_dict: dict[str,Any], output_keys: list[str],config: TrainConfig , **kwargs)-> Union[dict[str,Any], tuple, Any]:
        in_dim = input_dict['in_dim']
        out_dim = input_dict['out_dim']

        models = [self.get_model_by_config(model_config, in_dim, out_dim)  for model_config in config.model_configs]
        
        return models
    
    
    
    